@@ -26,6 +26,8 @@ AnyEvent::WebDriver - control browsers using the W3C WebDriver protocol
 
 =head1 DESCRIPTION
 
+WARNING: THE API IS NOT GUARANTEED TO BE STABLE UNTIL VERSION 1.0.
+
 This module aims to implement the W3C WebDriver specification which is the
 standardised equivalent to the Selenium WebDriver API., which in turn aims
 at remotely controlling web browsers such as Firefox or Chromium.
@@ -52,16 +54,15 @@ package AnyEvent::WebDriver;
 use common::sense;
 
 use Carp ();
-use JSON::XS ();
 use AnyEvent ();
 use AnyEvent::HTTP ();
 
-our $VERSION = 0.2;
+our $VERSION = 0.5;
 
 our $WEB_ELEMENT_IDENTIFIER = "element-6066-11e4-a52e-4f735466cecf";
 
-my $json = JSON::XS->new
-   ->utf8;
+my $json = eval { require JSON::XS; JSON::XS:: } || do { require JSON::PP; JSON::PP:: };
+$json = $json->new->utf8;
 
 $json->boolean_values (0, 1)
    if $json->can ("boolean_values");
@@ -114,7 +115,7 @@ sub AUTOLOAD {
    my $name_ = "$name\_";
 
    defined &$name_
-      or Carp::croak "AUTOLOAD: no such method";
+      or Carp::croak "$AUTOLOAD: no such method";
 
    my $func_ = \&$name_;
 
@@ -214,6 +215,59 @@ sub actions {
    AnyEvent::WebDriver::Actions->new (wd => $_[0])
 }
 
+=item $sessionstring = $wd->save_session
+
+Save the current session in a string so it can be restored load with
+C<load_session>. Note that only the session data itself is stored
+(currently the session id and capabilities), not the endpoint information
+itself.
+
+The main use of this function is in conjunction with disabled
+C<autodelete>, to save a session to e.g., and restore it later. It could
+presumably used for other applications, suhc as using the same sssion from
+multiple processes and so on.
+
+=item $wd->load_session ($sessionstring)
+
+=item $wd->set_session ($sessionid, $capabilities)
+
+Starts using the given session, as identified by
+C<$sessionid>. C<$capabilities> should be the original session
+capabilities, although the current version of this module does not make
+any use of it.
+
+The C<$sessionid> is stored in C<< $wd->{sid} >> (and could be fetched
+form there for later use), while the capabilities are stored in C<<
+$wd->{capabilities} >>.
+
+=cut
+
+sub save_session {
+   my ($self) = @_;
+
+   $json->encode ([1, $self->{sid}, $self->{capabilities}]);
+}
+
+sub load_session {
+   my ($self, $session) = @_;
+
+   $session = $json->decode ($session);
+
+   $session->[0] == 1
+      or Carp::croak "AnyEvent::WebDriver::load_session: session corrupted or from different version";
+
+   $self->set_session ($session->[1], $session->[2]);
+}
+
+sub set_session {
+   my ($self, $sid, $caps) = @_;
+
+   $self->{sid}          = $sid;
+   $self->{capabilities} = $caps;
+
+   $self->{_ep} = "$self->{endpoint}/session/$self->{sid}/";
+}
+
 =back
 
 =head2 SIMPLIFIED API
@@ -250,14 +304,54 @@ successfully, and only one session can be created per WebDriver object.
 On success, C<< $wd->{sid} >> is set to the session ID, and C<<
 $wd->{capabilities} >> is set to the returned capabilities.
 
+Simple example of creatring a WebDriver object and a new session:
+
    my $wd = new AnyEvent::Selenium endpoint => "http://localhost:4545";
+   $wd->new_session ({});
+
+Real-world example with capability negotiation:
 
    $wd->new_session ({
       capabilities => {
-         pageLoadStrategy => "normal",
-      }.
+         alwaysMatch => {
+            pageLoadStrategy        => "eager",
+            unhandledPromptBehavior => "dismiss",
+            # proxy => { proxyType => "manual", httpProxy => "1.2.3.4:56", sslProxy => "1.2.3.4:56" },
+         },
+         firstMatch => [
+            {
+               browserName => "firefox",
+               "moz:firefoxOptions" => {
+                  binary => "firefox/firefox",
+                  args => ["-devtools"],
+                  prefs => {
+                     "dom.webnotifications.enabled" => \0,
+                     "dom.disable_beforeunload" => \1,
+                     "browser.link.open_newwindow" => 3,
+                     "browser.link.open_newwindow.restrictions" => 0,
+                     "dom.popup_allowed_events" => "",
+                     "dom.disable_open_during_load" => \1,
+                  },
+               },
+            },
+            {
+               # generic fallback
+            },
+         ],
+
+      },
    });
 
+Firefox-specific capability documentation can be found L<on
+MDN|https://developer.mozilla.org/en-US/docs/Web/WebDriver/Capabilities>,
+Chrome-specific capability documentation might be found
+L<here|http://chromedriver.chromium.org/capabilities>, but the latest
+release at the time of this writing has effectively no WebDriver support
+at all, and canary releases are not freely downloadable.
+
+If you have URLs for Safari/IE/Edge etc. capabilities, feel free to tell
+me about them.
+
 =cut
 
 sub new_session_ {
@@ -267,12 +361,11 @@ sub new_session_ {
    $self->post_ (session => $kv, sub {
       my ($status, $res) = @_;
 
-      if ($status eq "200") {
-         $self->{sid}          = $res->{sessionId};
-         $self->{capabilities} = $res->{capabilities};
+      exists $res->{capabilities}
+         or $status = "500"; # blasted chromedriver
 
-         $self->{_ep} = "$self->{endpoint}/session/$self->{sid}/";
-      }
+      $self->set_session ($res->{sessionId}, $res->{capabilities})
+         if $status eq "200";
 
       $cb->($status, $res);
    });
@@ -408,7 +501,7 @@ sub close_window_ {
 }
 
 sub switch_to_window_ {
-   $_[0]->post_ (window => "$_[1]", $_[2]);
+   $_[0]->post_ (window => { handle => "$_[1]" }, $_[2]);
 }
 
 =item $handles = $wd->get_window_handles
@@ -492,11 +585,32 @@ sub fullscreen_window_ {
 
 =head3 ELEMENT RETRIEVAL
 
+To reduce typing and memory strain, the element finding functions accept
+some shorter and hopefully easier to remember aliases for the standard
+locator strategy values, as follows:
+
+   Alias   Locator Strategy
+   css     css selector
+   link    link text
+   substr  partial link text
+   tag     tag name
+
 =over
 
 =cut
 
-=item $element = $wd->find_element ($location_strategy, $selector)
+our %USING = (
+   css    => "css selector",
+   link   => "link text",
+   substr => "partial link text",
+   tag    => "tag name",
+);
+
+sub _using($) {
+   using => $USING{$_[0]} // "$_[0]"
+}
+
+=item $element = $wd->find_element ($locator_strategy, $selector)
 
 Finds the first element specified by the given selector and returns its
 element object. Raises an error when no element was found.
@@ -508,15 +622,15 @@ element object. Raises an error when no element was found.
    $element = $wd->find_element ("xpath" => '//input[@type="text"]');
    => e.g. { "element-6066-11e4-a52e-4f735466cecf" => "decddca8-5986-4e1d-8c93-efe952505a5f" }
 
-=item $elements = $wd->find_elements ($location_strategy, $selector)
+=item $elements = $wd->find_elements ($locator_strategy, $selector)
 
 As above, but returns an arrayref of all found element objects.
 
-=item $element = $wd->find_element_from_element ($element, $location_strategy, $selector)
+=item $element = $wd->find_element_from_element ($element, $locator_strategy, $selector)
 
 Like C<find_element>, but looks only inside the specified C<$element>.
 
-=item $elements = $wd->find_elements_from_element ($element, $location_strategy, $selector)
+=item $elements = $wd->find_elements_from_element ($element, $locator_strategy, $selector)
 
 Like C<find_elements>, but looks only inside the specified C<$element>.
 
@@ -530,19 +644,19 @@ Returns the active element.
 =cut
 
 sub find_element_ {
-   $_[0]->post_ (element => { using => "$_[1]", value => "$_[2]" }, $_[3]);
+   $_[0]->post_ (element => { _using $_[1], value => "$_[2]" }, $_[3]);
 }
 
 sub find_elements_ {
-   $_[0]->post_ (elements => { using => "$_[1]", value => "$_[2]" }, $_[3]);
+   $_[0]->post_ (elements => { _using $_[1], value => "$_[2]" }, $_[3]);
 }
 
 sub find_element_from_element_ {
-   $_[0]->post_ ("element/$_[1]/element" => { using => "$_[2]", value => "$_[3]" }, $_[4]);
+   $_[0]->post_ ("element/$_[1]/element" => { _using $_[2], value => "$_[3]" }, $_[4]);
 }
 
 sub find_elements_from_element_ {
-   $_[0]->post_ ("element/$_[1]/elements" => { using => "$_[2]", value => "$_[3]" }, $_[4]);
+   $_[0]->post_ ("element/$_[1]/elements" => { _using $_[2], value => "$_[3]" }, $_[4]);
 }
 
 sub get_active_element_ {
@@ -948,7 +1062,7 @@ entering some text and pressing enter, you can use this:
    $wd->actions
       ->click (1, 100)
       ->type ("some text")
-      ->key ("Enter")
+      ->key ("{Enter}")
       ->perform;
 
 By default, keyboard and mouse input sources are provided. You can create
@@ -958,13 +1072,9 @@ be more verbosely written like this:
    $wd->actions
       ->click (1, 100, "mouse")
       ->type ("some text")
-      ->key ("Enter")
+      ->key ("{Enter}")
       ->perform;
 
-   
-
-#TODO verboser example
-
 When you specify the event source expliticly it will switch the current
 "focus" for this class of device (all keyboards are in one class, all
 pointer-like devices such as mice/fingers/pens are in one class), so you
@@ -983,7 +1093,7 @@ package AnyEvent::WebDriver::Actions;
 =item $al = new AnyEvent::WebDriver::Actions
 
 Create a new empty action list object. More often you would use the C<<
-$sel->action_list >> method to create one that is already associated with
+$wd->action_list >> method to create one that is already associated with
 a given web driver.
 
 =cut
@@ -1016,7 +1126,7 @@ Example: switchdefine a new touch device called C<fatfinger>.
 
 =cut
 
-sub _default_source {
+sub _default_source($) {
    my ($source) = @_;
 
       $source eq "keyboard" ? { actions => [], id => $source, type => "key" }
@@ -1193,18 +1303,125 @@ sub move {
 
 Press or release the given key.
 
+=item $al = $al->key ($key, $source)
+
+Peess and release the given key, without unnecessary delay.
+
+A special syntax, C<{keyname}> can be used for special keys - all the special key names from
+L<section 17.4.2|https://www.w3.org/TR/webdriver1/#keyboard-actions> of the WebDriver recommendation
+can be used.
+
+Example: press and release "a".
+
+   $al->key ("a");
+
+Example: press and release the "Enter" key:
+
+   $al->key ("\x{e007}");
+
+Example: press and release the "enter" key using the special key name syntax:
+
+   $al->key ("{Enter}");
+
+=item $al = $al->type ($string, $source)
+
+Convenience method to simulate a series of key press and release events
+for the keys in C<$string>. There is no syntax for special keys,
+everything will be typed "as-is" if possible.
+
 =cut
 
+our %SPECIAL_KEY = (
+   "Unidentified"   => 0xE000,
+   "Cancel"         => 0xE001,
+   "Help"           => 0xE002,
+   "Backspace"      => 0xE003,
+   "Tab"            => 0xE004,
+   "Clear"          => 0xE005,
+   "Return"         => 0xE006,
+   "Enter"          => 0xE007,
+   "Shift"          => 0xE008,
+   "Control"        => 0xE009,
+   "Alt"            => 0xE00A,
+   "Pause"          => 0xE00B,
+   "Escape"         => 0xE00C,
+   " "              => 0xE00D,
+   "PageUp"         => 0xE00E,
+   "PageDown"       => 0xE00F,
+   "End"            => 0xE010,
+   "Home"           => 0xE011,
+   "ArrowLeft"      => 0xE012,
+   "ArrowUp"        => 0xE013,
+   "ArrowRight"     => 0xE014,
+   "ArrowDown"      => 0xE015,
+   "Insert"         => 0xE016,
+   "Delete"         => 0xE017,
+   ";"              => 0xE018,
+   "="              => 0xE019,
+   "0"              => 0xE01A,
+   "1"              => 0xE01B,
+   "2"              => 0xE01C,
+   "3"              => 0xE01D,
+   "4"              => 0xE01E,
+   "5"              => 0xE01F,
+   "6"              => 0xE020,
+   "7"              => 0xE021,
+   "8"              => 0xE022,
+   "9"              => 0xE023,
+   "*"              => 0xE024,
+   "+"              => 0xE025,
+   ","              => 0xE026,
+   "-"              => 0xE027,
+   "."              => 0xE028,
+   "/"              => 0xE029,
+   "F1"             => 0xE031,
+   "F2"             => 0xE032,
+   "F3"             => 0xE033,
+   "F4"             => 0xE034,
+   "F5"             => 0xE035,
+   "F6"             => 0xE036,
+   "F7"             => 0xE037,
+   "F8"             => 0xE038,
+   "F9"             => 0xE039,
+   "F10"            => 0xE03A,
+   "F11"            => 0xE03B,
+   "F12"            => 0xE03C,
+   "Meta"           => 0xE03D,
+   "ZenkakuHankaku" => 0xE040,
+   "Shift"          => 0xE050,
+   "Control"        => 0xE051,
+   "Alt"            => 0xE052,
+   "Meta"           => 0xE053,
+   "PageUp"         => 0xE054,
+   "PageDown"       => 0xE055,
+   "End"            => 0xE056,
+   "Home"           => 0xE057,
+   "ArrowLeft"      => 0xE058,
+   "ArrowUp"        => 0xE059,
+   "ArrowRight"     => 0xE05A,
+   "ArrowDown"      => 0xE05B,
+   "Insert"         => 0xE05C,
+   "Delete"         => 0xE05D,
+);
+
+sub _kv($) {
+   $_[0] =~ /^\{(.*)\}$/s
+      ? (exists $SPECIAL_KEY{$1}
+           ? chr $SPECIAL_KEY{$1}
+           : Carp::croak "AnyEvent::WebDriver::Actions: special key '$1' not known")
+      : $_[0]
+}
+
 sub key_down {
    my ($self, $key, $source) = @_;
 
-   $self->_add ($source, kbd => keyDown => value => $key)
+   $self->_add ($source, kbd => keyDown => value => _kv $key)
 }
 
 sub key_up {
    my ($self, $key, $source) = @_;
 
-   $self->_add ($source, kbd => keyUp => value => $key)
+   $self->_add ($source, kbd => keyUp => value => _kv $key)
 }
 
 sub key {
@@ -1215,6 +1432,15 @@ sub key {
       ->key_up   ($key)
 }
 
+sub type {
+   my ($self, $string, $source) = @_;
+
+   $self->key ($_, $source)
+      for $string =~ /(\X)/g;
+
+   $self
+}
+
 =item $al->perform ($wd)
 
 Finaluses and compiles the list, if not done yet, and calls C<<
